@@ -17,6 +17,7 @@ import logging
 import time
 import threading
 
+from botocore.vendored.requests.adapters import HTTPAdapter
 from botocore.vendored.requests.sessions import Session
 from botocore.vendored.requests.utils import get_environ_proxies
 from botocore.vendored.requests.exceptions import ConnectionError
@@ -29,12 +30,15 @@ from botocore.exceptions import ConnectionClosedError
 from botocore.compat import filter_ssl_warnings
 from botocore.utils import is_valid_endpoint_url
 from botocore.hooks import first_non_none_response
+from botocore.history import get_global_history_recorder
 from botocore.response import StreamingBody
 from botocore import parsers
 
 
 logger = logging.getLogger(__name__)
+history_recorder = get_global_history_recorder()
 DEFAULT_TIMEOUT = 60
+MAX_POOL_CONNECTIONS = 10
 filter_ssl_warnings()
 
 try:
@@ -63,19 +67,43 @@ def convert_to_response_dict(http_response, operation_model):
     response_dict = {
         'headers': http_response.headers,
         'status_code': http_response.status_code,
+        'context': {
+            'operation_name': operation_model.name,
+        }
     }
     if response_dict['status_code'] >= 300:
         response_dict['body'] = http_response.content
+    elif operation_model.has_event_stream_output:
+        response_dict['body'] = http_response.raw
     elif operation_model.has_streaming_output:
-        response_dict['body'] = StreamingBody(
-            http_response.raw, response_dict['headers'].get('content-length'))
+        length = response_dict['headers'].get('content-length')
+        response_dict['body'] = StreamingBody(http_response.raw, length)
     else:
         response_dict['body'] = http_response.content
     return response_dict
 
 
-class PreserveAuthSession(Session):
+class BotocoreHTTPSession(Session):
+    """Internal session class used to workaround requests behavior.
+
+    This class is intended to be used only by the Endpoint class.
+
+    """
+    def __init__(self, max_pool_connections=MAX_POOL_CONNECTIONS,
+                 http_adapter_cls=HTTPAdapter):
+        super(BotocoreHTTPSession, self).__init__()
+        # In order to support a user provided "max_pool_connections", we need
+        # to recreate the HTTPAdapter and pass in our max_pool_connections
+        # value.
+        adapter = http_adapter_cls(pool_maxsize=max_pool_connections)
+        # requests uses an HTTPAdapter for mounting both http:// and https://
+        self.mount('https://', adapter)
+        self.mount('http://', adapter)
+
     def rebuild_auth(self, prepared_request, response):
+        # Keep the existing auth information from the original prepared request.
+        # Normally this method would be where auth is regenerated as needed.
+        # By making this a noop, we're keeping the existing auth info.
         pass
 
 
@@ -92,7 +120,8 @@ class Endpoint(object):
 
     def __init__(self, host, endpoint_prefix,
                  event_emitter, proxies=None, verify=True,
-                 timeout=DEFAULT_TIMEOUT, response_parser_factory=None):
+                 timeout=DEFAULT_TIMEOUT, response_parser_factory=None,
+                 max_pool_connections=MAX_POOL_CONNECTIONS):
         self._endpoint_prefix = endpoint_prefix
         self._event_emitter = event_emitter
         self.host = host
@@ -100,8 +129,10 @@ class Endpoint(object):
         if proxies is None:
             proxies = {}
         self.proxies = proxies
-        self.http_session = PreserveAuthSession()
+        self.http_session = BotocoreHTTPSession(
+            max_pool_connections=max_pool_connections)
         self.timeout = timeout
+        self.max_pool_connections = max_pool_connections
         logger.debug('Setting %s timeout as %s', endpoint_prefix, self.timeout)
         self._lock = threading.Lock()
         if response_parser_factory is None:
@@ -174,9 +205,20 @@ class Endpoint(object):
         # If no exception occurs then exception is None.
         try:
             logger.debug("Sending http request: %s", request)
+            history_recorder.record('HTTP_REQUEST', {
+                'method': request.method,
+                'headers': request.headers,
+                'streaming': operation_model.has_streaming_input,
+                'url': request.url,
+                'body': request.body
+            })
+            streaming = any([
+                operation_model.has_streaming_output,
+                operation_model.has_event_stream_output
+            ])
             http_response = self.http_session.send(
                 request, verify=self.verify,
-                stream=operation_model.has_streaming_output,
+                stream=streaming,
                 proxies=self.proxies, timeout=self.timeout)
         except ConnectionError as e:
             # For a connection error, if it looks like it's a DNS
@@ -201,12 +243,18 @@ class Endpoint(object):
                          exc_info=True)
             return (None, e)
         # This returns the http_response and the parsed_data.
-        response_dict = convert_to_response_dict(http_response,
-                                                 operation_model)
-        parser = self._response_parser_factory.create_parser(
-            operation_model.metadata['protocol'])
+        response_dict = convert_to_response_dict(http_response, operation_model)
+
+        http_response_record_dict = response_dict.copy()
+        http_response_record_dict['streaming'] = \
+            operation_model.has_streaming_output
+        history_recorder.record('HTTP_RESPONSE', http_response_record_dict)
+
+        protocol = operation_model.metadata['protocol']
+        parser = self._response_parser_factory.create_parser(protocol)
         parsed_response = parser.parse(
             response_dict, operation_model.output_shape)
+        history_recorder.record('PARSED_RESPONSE', parsed_response)
         return (http_response, parsed_response), None
 
     def _looks_like_dns_error(self, e):
@@ -241,16 +289,22 @@ class EndpointCreator(object):
 
     def create_endpoint(self, service_model, region_name, endpoint_url,
                         verify=None, response_parser_factory=None,
-                        timeout=DEFAULT_TIMEOUT):
+                        timeout=DEFAULT_TIMEOUT,
+                        max_pool_connections=MAX_POOL_CONNECTIONS,
+                        proxies=None):
         if not is_valid_endpoint_url(endpoint_url):
+
             raise ValueError("Invalid endpoint: %s" % endpoint_url)
+        if proxies is None:
+            proxies = self._get_proxies(endpoint_url)
         return Endpoint(
             endpoint_url,
             endpoint_prefix=service_model.endpoint_prefix,
             event_emitter=self._event_emitter,
-            proxies=self._get_proxies(endpoint_url),
+            proxies=proxies,
             verify=self._get_verify_value(verify),
             timeout=timeout,
+            max_pool_connections=max_pool_connections,
             response_parser_factory=response_parser_factory)
 
     def _get_proxies(self, url):
